/*

----------------------------------------------------------
Makro: MODEL_CLUSTER1
Nzev: Shlukuje pozorovn s ohledem na spojitou promnnou
Verze: 2.3
Autor: jan.ruzicka@gmail.com
----------------------------------------------------------

-----
POPIS
-----

Shlukuje pozorovn z tabulky DATA= s ohledem na spojitou promnnou (VAR=).
Pozorovn jsou urena promnnou ID=, etnost (vha) kadho pozorovn je urena
promnnou FREQ= (pedpokld se, e data jsou pedsumarizovna).

Je mon vybrat ze dvou typ algoritm podle typu ID promnn. TYPE=INTERVAL provede
shlukovn s ohledem na ID= i VAR= promnn (hod se pro spojitou ID promnnou).
TYPE=NOMINAL shlukuje pouze podel VAR=, je defaultn nastaven.

Defaultn nastaven metody je METHOD=HIERARCHICAL, kernelov shlukovn zajistme
pomoc METHOD=KERNEL, k-means algoritmus zskme pomoc volby METHOD=KMEANS, piem
tato metoda je urena hlavn na rozshl data, ale pozor, neprovd standardizaci.
Maximln poet vytvoench shluk je mon omezit (volba MAXCLUSTERS=).

Pro METHOD=HIERARCHICAL je mon volit ze dvou algoritm vbru optimlnho potu
shluk (NCMETHOD=1 je defaultn nastaven), piem NCMETHOD=1 vede k vbru vtho
potu shluk.

Vstup:

data - tabulka uren parametrem DATA=
     - mus obsahovat promnn definovan parametry ID=,VAR= a nepovinn (FREQ=)

Vstup:

out        - tabulka obsahujc popis shluk
fmtout.out - formt pevdjc promnnou ID na slo shluku (pokud je poadovn)
           - msto sla shluku me formtovan hodnota bt popisem shluku i slem intervalu
           - dendrogram popisujc segmenty graficky (pokud je poadovn)
           - XY graf s berevnm vyznaenm shluk

------
SYNTAX
------

%model_cluster1
(
    data=,               *** vstupn dataset
    id=,                 *** ID shlukovanch pozorovn
    var=,                *** podle jak promnn shlukovat
    freq=,               *** ve kter promnn je uloena vha
    type=nominal,        *** typ ID (nominln, intervalov)
    method=hierarchical, *** metoda shlukovn (hierarchick, kernelov, k-means)
    ncmethod=,           *** metoda na vbr optimlnho potu shluk (1,2)
    maxclusters=,        *** maximln poet shluk
    out=,                *** vstupn tabulka s popisem shluk (nen-li, pouije se ID=)
    fmtout=,             *** knihovna i katalog, kam uloit vstupn formt
    fmtname=,            *** nzev vstupnho formtu (nen-li, pouije se ID=)
    fmtvalue=            *** formtovan hodnota (cluster, order, interval) 
    graph=all)           *** grafick vstupy (dn, dendrogram, XY graf, vechny)
)

-------------
HISTORIE ZMN
-------------

21.05.2002  Verze 1.0
24.05.2002  Verze 1.1
            + pidna volba METHOD=
29.05.2002  Verze 1.2
            + pidna volba NCMETHOD=
24.09.2002  Verze 2.0
            - pravy algoritmu
01.11.2003  Verze 2.20
            - zmna nzvu makra
28.02.2009  Verze 2.3

--------
PKLADY
--------

*** Prmrn vky (HEIGHT) podle vku (AGE) uloit do tabulky CLASS;

proc summary data=sashelp.class nway;
class age;
output out=class mean(height)=height;
run;

* shlukovn tchto vkovch skupin do interval (TYPE=INTERVAL) s ohledem na velikost HEIGHT;
* FREQ=_freq_ definuje promnnou, obsahujc poet pozorovn, kter reprezentuj skupinu AGE;
* vstupem bude tabulka WORK.HEIGHT a formt HEIGHT. uloen v katalogu WORK.FORMATS;

%model_cluster1
(
    data=class,
    id=age,
    var=height,
    freq=_freq_,
    type=interval,
    out=my.height,
    fmtout=my.formats
)

* shlukovn jmen (promnn NAME) z tabulky SASHELP.CLASS s ohledem na velikost WEIGHT;
* bude pouita defaultn Greenacrova metoda (TYPE=NOMINAL), protoe NAME je nominln promnn;
* vstupem bude tabulka WEIGHT a oba typy grafickch vstup (dendrogram a XY graf);

%model_cluster1
(
    data=sashelp.class,
    id=name,
    var=weight,
    ncmethod=2,
    out=my.weight,
    graph=all
)

--------
POZNMKY
--------

Vyaduje modul SAS/STAT a voliteln SAS/GRAPH pro grafick vstupy.
Pipomnky a npady k vylepen vtny.

*/

%macro model_cluster1
(
    data=,
    id=,
    var=,
    freq=,
    type=nominal,
    method=hierarchical,
    ncmethod=,
    maxclusters=,
    out=,
    fmtout=,
    fmtname=,
    fmtvalue=,
    graph=all
) / des='Categorizes variable and creates format (clustering method)';

%put;
%put NOTE: Macro MODEL_CLUSTER1 2.3;
%put;

%local i dupid ncl rsq dsid idtype idfmt vartype pchi fmtname outlib;
%let i=0;

%if %sysfunc(sysprod(SAS/STAT))=-1 %then %do;
    %put ERROR: SAS/STAT licence is required.;
    %goto mexit;
%end;

%if &data= %then %do;
    %put ERROR: Input dataset is required (DATA=).;
    %goto mexit;
%end;

%if &id= %then %do;
    %put ERROR: ID variable is required (ID=).;
    %goto mexit;
%end;
%else %if %scan(&id,2) ne %then %do;
    %put ERROR: Only one ID variable is required.;
    %goto mexit;
%end;

%if &var= %then %do;
    %put ERROR: Target variable is required (VAR=).;
    %goto mexit;
%end;
%else %if %scan(&var,2) ne %then %do;
    %put ERROR: Only one target variable is required.;
    %goto mexit;
%end;

%let dsid=%sysfunc(open(&data));
%if &dsid %then %do;
    %if %sysfunc(varnum(&dsid,&id))=0 %then %do;
        %put ERROR: ID variable %upcase(&id) is not available.;
        %goto mexit;
    %end;
    %else %do;
        %let idtype=%sysfunc(vartype(&dsid,%sysfunc(varnum(&dsid,&id))));
        %let idfmt=%sysfunc(varfmt(&dsid,%sysfunc(varnum(&dsid,&id))));
    %end;
    %if %sysfunc(varnum(&dsid,&var)) %then %do;
        %let vartype=%sysfunc(vartype(&dsid,%sysfunc(varnum(&dsid,&var))));
        %if &vartype=C %then %do;
            %put ERROR: Target variable %upcase(&var) is not numeric.;
            %goto mexit;
        %end;
    %end;
    %else %do;
        %put ERROR: Target variable %upcase(&var) is not available.;
        %goto mexit;
    %end;
    %let dsid=%sysfunc(close(&dsid));
%end;
%else %do;
    %put ERROR: Input dataset %upcase(&data) is not available.;
    %goto mexit;
%end;

%let type=%upcase(&type);
%if &type ne NOMINAL %then %do;
    %if &type ne INTERVAL %then %do;
        %put ERROR: Clustering type must be either INTERVAL or NOMINAL.;
        %goto mexit;
    %end;
    %else %if &idtype=C %then %do;
        %put ERROR: ID variable must be numeric for INTERVAL clustering.;
        %goto mexit;
    %end;
%end;

%if %sysfunc(translate(1&maxclusters,"          ","0123456789")) ne %then %do;
    %put ERROR: MAXCLUSTERS= option expects integer number.;
    %goto mexit;
%end;
%if %scan(&maxclusters,2) ne %then %do;
    %put ERROR: MAXCLUSTERS= option expects only one number.;
    %goto mexit;
%end;
%if &maxclusters ne %then %if &maxclusters<1 %then %do;
    %put ERROR: MAXCLUSTERS= option expects number >=1.;
    %goto mexit;
%end;

%let method=%upcase(&method);
%if &method ne HIERARCHICAL %then %if &method ne KERNEL %then %if &method ne KMEANS %then %do;
    %put ERROR: Clustering method must be HIERARCHICAL, KERNEL or KMEANS.;
    %goto mexit;
%end;
%if &method=KMEANS %then %if &maxclusters= %then %do;
    %put ERROR: MAXCLUSTERS= must be defined when METHOD=KMEANS.;
    %goto mexit;
%end;

%if &method=HIERARCHICAL %then %do;
    %if &ncmethod= %then %let ncmethod=1;
    %if &ncmethod ne 1 %then %if &ncmethod ne 2 %then %do;
        %put ERROR: Valid values for NCMETHOD= option are 1 or 2.;
        %goto mexit;
    %end;
%end;
%else %if &ncmethod ne %then %do;
    %put ERROR: NCMETHOD= option is valid only for METHOD=HIERARCHICAL.;
    %goto mexit;
%end;

%if &freq ne %then %if %scan(&freq,2) ne %then %do;
    %put ERROR: Only one frequency variable could be defined.;
    %goto mexit;
%end;

%if &out= %then %do;
    %let out=&id;
    %put NOTE: Output table name will be WORK.%upcase(&id).;
%end;
%if %scan(&out,2,.)= %then %let outlib=WORK;
%else %do;
    %let outlib=%scan(&out,1,.);
    %let out=%scan(&out,2,.);
%end;

%if &fmtname= %then %do;
    %let fmtname=&id;
    %let i=1;
%end;
%let fmtname=%upcase(&fmtname);
%if &idtype=C %then %do;
    %if %length(&fmtname)>7 %then %let i=1;
    %let fmtname=%substr(&fmtname,1,%sysfunc(min(7,%length(&fmtname))));
    %if &i %then %if &fmtout ne %then %put NOTE: Output format name will be $&fmtname..;
%end;
%else %do;
    %if %length(&fmtname)>8 %then %let i=1;
    %let fmtname=%substr(&fmtname,1,%sysfunc(min(8,%length(&fmtname))));
    %if &i %then %if &fmtout ne %then %put NOTE: Output format name will be &fmtname..;
%end;

%let graph=%upcase(&graph);
%if &graph ne NO %then %if &graph ne TREE %then %if &graph ne PLOT %then %if &graph ne ALL %then %do;
    %put ERROR: Valid values for GRAPH= option are NO, TREE, PLOT or ALL.;
    %goto mexit;
%end;
%if &method ne HIERARCHICAL %then %if &graph=TREE %then %do;
    %put ERROR: GRAPH=TREE option is only when METHOD=HIERARCHICAL.;
    %goto mexit;
%end;
%if &graph ne NO %then %if %sysfunc(sysprod(SAS/GRAPH))=-1 %then %do;
    %put WARNING: SAS/GRAPH licence is required to produce graphic output.;
    %let graph=NO;
    %goto mexit;
%end;

%let fmtvalue=%upcase(&fmtvalue);
%if &fmtvalue ne %then %if &fmtout= %then %do;
    %put ERROR: FMTVALUE= option is valid only with FMTOUT= defined.;
    %goto mexit;
%end;
%if &fmtvalue= %then %let fmtvalue=CLUSTER;
%if &fmtvalue ne CLUSTER %then %if &fmtvalue ne ORDER %then %if &fmtvalue ne INTERVAL %then %do;
    %put ERROR: Valid values for FMTVALUE= option are CLUSTER, ORDER or INTERVAL.;
    %goto mexit;
%end;
%if &fmtvalue ne CLUSTER %then %if &type ne INTERVAL %then %do;
    %put ERROR: Only FMTVALUE=CLUSTER is valid when TYPE=NOMINAL.;
    %goto mexit;
%end;

%if &method=HIERARCHICAL %then %do;

proc sql outobs=1 noprint;
select distinct count(&id)>1 as dup into :dupid
    from &data (keep=&id) group by &id order by dup descending;
quit;

%if &dupid %then %do;
    %put ERROR: Non-distinct ID found;
    %goto mexit;
%end;

proc freq data=&data noprint;
tables &id*&var / chisq;
output out=_temp (keep=n _pchi_) chisq;
run;

data _null_;
set _temp;
if n<=2 then call symput('pchi','error');
else call symput('pchi',compress(put(_pchi_,best.)));
run;

%if &pchi=error %then %do;
    proc datasets lib=work mt=data nolist;
    delete _temp;
    quit;
    %put ERROR: Variable ID has less than 3 levels;
    %goto mexit;
%end;

ods listing close;
ods output clusterhistory=_&out (keep=numberofclusters rsquared);
proc cluster data=&data method=ward
             outtree=_outtree (keep=&id _name_ _parent_ _rsq_ _ncl_)
             %if &type=INTERVAL %then standard;;
%if &freq ne %then freq &freq%str(;);
%if &type=NOMINAL %then var &var%str(;);
                  %else var &id &var%str(;);
id &id;
run;
ods output close;
ods listing;

data _null_;
set _&out %if &maxclusters ne %then (where=(numberofclusters<=&maxclusters)); end=_finish_;
retain _min_ _optrsq_;
%if &ncmethod=1 %then %do;
    if _n_=1 then do;
        set _temp;
        _min_=numberofclusters;
        _optrsq_=rsquared;
    end;        
    _lagrsq_=lag(rsquared);
    if not _finish_ then do;
        _logpvalue_=logsdf('chisq',&pchi*rsquared,numberofclusters-1);
        if _logpvalue_<lag(_logpvalue_) then do;
            _min_=numberofclusters;
            _optrsq_=_lagrsq_;
        end;
    end;
%end;

%else %do;
    if _n_=1 then do;
        _min_=numberofclusters;
        _optrsq_=rsquared;
    end;
    _lagrsq_=lag(rsquared);
    _r2dif_=abs(dif(dif(rsquared*100)));
    if _r2dif_>lag(_r2dif_) and numberofclusters>1 then do;
        _min_=numberofclusters+1;
        _optrsq_=_lagrsq_;
    end;
%end;

else do;
    call symput('ncl',compress(put(_min_,best.)));
    call symput('rsq',compress(put(_optrsq_,best.)));
end;
run;

proc tree data=_outtree nclusters=&ncl out=_&out (keep=&id cluster)
%if &graph ne NO %then %if &graph ne PLOT %then h=rsq horizontal maxheight=&rsq;
%if &graph ne ALL %then %if &graph ne TREE %then noprint;;
id &id;
run;

%if &graph ne NO %then %if &graph ne TREE %then %do;
    proc sql;
    create view _plot as
    select a.&id,a.&var,b.cluster from &data as a full join _&out as b on
    %if &idtype=N %then round(a.&id)=round(b.&id);
    %else a.&id=b.&id;;
    quit;
    symbol1 v=dot h=1;
    proc gplot data=_plot;
    plot &var*&id=cluster / grid;
    run;
    quit;
%end;

%end;
%else %do;

%if &method=KERNEL %then %do;
proc modeclus data=&data noprint out=&outlib..&out (keep=&id &var cluster)
              %if &type=INTERVAL %then standard;
              %if &maxclusters ne %then maxclusters=&maxclusters m=6;%else m=1;;
%if &freq ne %then freq &freq%str(;);
id &id;
%if &type=NOMINAL %then var &var%str(;);
                  %else var &id &var%str(;);
id &id;
run;
%end;
%else %do;

proc fastclus data=&data noprint out=&outlib..&out (keep=&id &var cluster) maxclusters=&maxclusters;;
%if &freq ne %then freq &freq%str(;);
%if &type=NOMINAL %then var &var%str(;);
                  %else var &id &var%str(;);
id &id;
run;
%end;

%if &graph ne NO %then %do;
    symbol1 v=dot h=1;
    proc gplot data=&outlib..&out;
    plot &var*&id=cluster / grid;
    run;
    quit;
%end;

%end;

proc sort data=%if &method=HIERARCHICAL %then _&out;%else &outlib..&out;
          (keep=&id cluster rename=(cluster=_cluster_))
      %if &type=INTERVAL %then out=&outlib..&out (rename=(&id=&id._start));
                         %else out=&outlib..&out;;
by &id%if &type=INTERVAL %then _cluster_;;
run;

%if &type=NOMINAL %then %do;
    data &outlib..&out (keep=&id _cluster_ sortedby=&id)
    %if &fmtout ne %then _&out (rename=(&id=start _&fmtvalue._=label)
                                keep=&id _&fmtvalue._ fmtname type);;
    attrib &id label='';
%end;
%else %do;
    data &outlib..&out (keep=&id._start &id._end _cluster_ _interval_ _order_ sortedby=&id._start)
    %if &fmtout ne %then _&out (rename=(&id._start=start &id._end=end _&fmtvalue._=label)
                                keep=&id._start &id._end _&fmtvalue._ fmtname type);;
    attrib _order_ label='' &id._start label='' &id._end label='';
    %if &idfmt ne %then format &id._end &idfmt%str(;);
%end;

set &outlib..&out end=_finish_;
%if &type=INTERVAL %then %do;
    by _cluster_ notsorted;
    retain _start_;
%end;

%if &fmtout ne %then %do;
%if &idtype=C %then type='C';
              %else type='N';;
fmtname="&fmtname";
%end;

%if &type=INTERVAL %then %do;
    if first._cluster_ then _start_=&id._start;
    if last._cluster_ then do;
        &id._end=&id._start;
        &id._start=_start_;
        if &id._start ne &id._end then
        %if &idfmt ne %then %do;
            _interval_=compress(put(&id._start,&idfmt))||' - '||compress(put(&id._end,&idfmt));
            else _interval_=compress(put(&id._start,&idfmt));
        %end;
        %else %do;
            _interval_=compress(put(&id._start,best.))||' - '||compress(put(&id._end,best.));
            else _interval_=compress(put(&id._start,best.));
        %end;
        _order_+1;
        output;
    end;
%end;
run;

%if &fmtout ne %then %do;
    proc format lib=&fmtout cntlin=_&out;
    run;
%end;

%if &method=HIERARCHICAL %then %do;
proc datasets lib=work mt=(data view) nolist;
delete _outtree _temp _&out
       %if &graph ne NO %then %if &graph ne TREE %then _plot;;
quit;
%end;
%else %if &fmtout ne %then %do;
proc datasets lib=work mt=data nolist;
delete _&out;
quit;
%end;

%mexit:

%mend model_cluster1;
