/*

--------------------------------------------------------------------------------
Makro: MODEL_HIER1
Nzev: Shlukuje pozorovn s ohledem na spojitou prominnou a definovan pravidla
Verze: 2.3
Autor: jaroslav.pulpan@cze.sas.com
       jan.ruzicka@gmail.com
--------------------------------------------------------------------------------

-----
POPIS
-----

Clustering procedure based on one categorical variable, distances are calculated from single
continuous variable hierarchical clustering method is applied to levels (categories, bins)
of the input categorical variable.

Main difference between this procedure and e.g. PROC CLUSTER is restriction of levels that
could be joined. Restriction is defined by type of categorical variable that defines levels:

Ordinal category            - any level could be joined only with next or previous levels
Nominal category            - any level could be collapsed with any other (like PROC CLUSTER)
Restricted nominal category - only levels defined by adjacency dataset could be collapsed
Map category                - adjacency of levels is defived directly from SAS map dataset
                              adjacency datasets define matrix which is updated within in each
                              step of clustering procedure, updates is analogical to joining
                              of the areas on map (!!! yes this was the main goal of this macro)

Input:      DATA   - name of the summarized dataset with categories and segmentation variable*
            ID     - variable that defines categories
            VAR    - continuous variable that defines distance
            WEIGHT - weight variable 

Output:     set of datasets with cluster definitnion for each step
            (restricted by MINCLUSTERS and MAXCLUSTERS)
            set of HTML pages defined by HTML* macro parameters, it includes dendogram, tables,
            graphs and maps for each step within MINCLUSTERS and MAXCLUSTERS),
            animated GIF for map category

------
SYNTAX
------

%macro model_hier1
(
    data=,          *** name of the summarized dataset with category and segmentation variable
    id=,            *** name of the variable with category to be segmented
    type=1,         *** adjacency definition for category variable ID - it specifies which levels
                        of ID could be collapsed (0-2 ordinal, 3 nominal, 4 restricted nominal, 5 map)
                        0 - unformatted ordinal category variable (level could be collapsed with next or
                            previous, order is defined by unformatted values of class levels)
                        1 - formatted ordinal category variable (level could be collapsed with next or
                            previous, order is defined by formatted values of class levels)
                        2 - ordinal category variable with specific sort order (level could be collapsed
                            with next or previous, order is defined by dataset provided as ADJACENCY parameter)
                        3 - nominal ID category variable (any level could be collapsed with any
                            other, by using "Kratochvil's lemma" macro converts nominal category to
                            kind of ordinal variable for which order of levels is defined by of their
                            corresponding Y values for which level could be collapsed with next or previous)
                        4 - category variable with only limited set of allowed adjacencies, adjacency
                            is defined by ADJACENCY dataset, after joining levels L1 and L2 into new segment LX,
                            it is allowed to join it with all levels that were neighbouring to L1 and L2,
                            if map parameter is present maps graphical outputs are produced
                        5 - analogy of TYPE=4 only adjacency dataset is directly derived from map dataset
                            specified as MAP parameter ADJACENCY dataset is generated from the map
                            (in case, that ADJACENCY dataset &MAP_join doesnt exist)
    var=,           *** name of the variable Y that direct segmentation
    weight=,        *** weight variable

    adjacency=,     *** the name of the dataset that defines adjacency, it is valid for TYPE=2,4,5
                         dataset format for TYPE=2 - dataset with sorted lists of levels
                                                     (only necessary column is : ID)
                         dataset format for TYPE=4 - dataset with allowed adjoin levels defined
                                                     by its rows of form: ID_1 ID_2
                         dataset format for TYPE=5 - this dataset could be generated
                                                     (format will be same as for TYPE=4)
    predefined=,    *** name of the dataset with predefined set of clusters ;
                        dataset has form ID_1 ID_2, at the beginnig of the clustering procedure,
                        all levels defined in this dataset will be joined and then procedure will
                        continue driven by clusters defined by PREDEFINED=
    map=,           *** name of the map dataset to be used for plottinng of the maps ;
                        map dataset should contain ID as valid identification variable for the map,
                        plus X,Y vars that defines coordinates of the polygons on map

    minclusters=,                            *** minimun number of clusters for which detail result
                                                 dataset will be generated/plotted ; default is 1
    maxclusters=,                            *** maximal number of clusters for which detail result
                                                 dataset will be generated/plotted ; default is
                                                 min((initial number of levels of ID)-1,300)

    outlib=work,                             *** name of the library for output datasets

    plots_only=no,                           *** switch that requests only generation of plots (YES|NO)
                                                 (clustering should be run beforehand)
    rep_root=c:\temp,                        *** path of the HTML output
    rep_css=../all_reports.css,              *** specif. CSS
    rep_all=all_reports.html,                *** specif. root HTML

    animgif=&rep_root\animgif.gif,           *** name of the animated GIF file for map dataset
    colorscheme=,                            *** color scheme: empty - by Y values,
                                                               1 - by number of NG levels
    title=Segmentation of &id based on &var, *** title to be defined on output
    title2=,
    footnote=Based on dataset &data,         *** footnote to be defined on output
    nlevels_detail=200,                      *** max. number of obs. for which detail list of
                                                 levels is printed and HTML flyover are generated
    nlevels_axis=20,                         *** max. number of obs. for which detail axis is prepared
    log_id=no,                               *** generate LOG scale on ID axis (YES|NO)
    log_var=no,                              *** generate LOG scale on VAR axis (YES|NO)
    log_weight=no,                           *** generate LOG scale on WEIGHT axis (YES|NO)
                                                 displays category level scale, that is
                                                 1. LOG_ID=YES this is valid only for ADJCAT<3
                                                 2. use just for numerical ID
    bubble_plot=1,                           *** switch that generates bubble (circles) by weight
                                                 variable in Y variable graphs
    bubble_size=1,                           *** normalization coeff. for buble size change is
                                                 there are too much overlaps
    weight_plot=0,                           *** switch that generates weight plots,
                                                 (by default it is set by bubble_plot switch)
                                                 set to 1 to produce them, 0 to supress them
    debug=no,                                *** debug request flag (YES/NO)
)

-------------
HISTORIE ZMIN
-------------

06/2002     Initial coding
06/2002     change of the minimalization criterium to sum of squares
08/2002     dendograms included to output
09/2002     rewritten from adj_cluster1 to include class adjacency defined by map
10/2002     parameters definiton changed/animated maps added 
11/2002     graphival options enriched (log scale, bubles, no weight)
12/2002     predefined joins added
01.11.2003  Verze 2.20
            + zapracovn do jednotnho frameworku SAS maker
            - zmna nzvu makra
28.02.2009  Verze 2.3

------------------
PLNY NA VYLEPEN
------------------

Zminit vykreslovn map tak, aby fungovaly HTML linky na Prahu (jako ostruvek ve STE kraji).

Z WEIGHT= udilat nepovinn parametr.

--------
POKLADY
--------

*** Prumirn vky (HEIGHT) podle viku (AGE) uloit do tabulky CLASS;

proc summary data=sashelp.class nway;
class age;
output out=class mean(height)=height;
run;

* shlukovn tichto vikovch skupin do intervalu (TYPE=INTERVAL) s ohledem na velikost HEIGHT;
* FREQ=_freq_ definuje prominnou, obsahujc poeet pozorovn, kter reprezentuj skupinu AGE;
* vstupem bude tabulka WORK.HEIGHT a formt HEIGHT. uloen v katalogu WORK.FORMATS;

%model_hier1
(
    data=class,
    id=age,
    var=height,
    weight=_freq_,
    type=1
)

* shlukovn jmen (prominn NAME) z tabulky SASHELP.CLASS s ohledem na velikost WEIGHT;
* bude pouita defaultn Greenacrova metoda (TYPE=NOMINAL), protoe NAME je nominln prominn;
* vstupem bude tabulka WEIGHT a oba typy grafickch vstupu (dendrogram a XY graf);

%model_hier1
(
    data=sashelp.class,
    id=name,
    var=weight
)

--------
POZNMKY
--------

Vyaduje volitelni modul SAS/GRAPH pro grafick vstupy. Vyuv nsledujc makra pro:

%VAR_FORMAT - zjitin formtu prominn
%VAR_LENGTH - zjitin dlky prominn
%VAR_LABEL  - zjitin labelu prominn
%DS_VARS    - testovn struktury vstupnho datasetu

Poipomnky a npady k vylepen vtny.

*/

%macro model_hier1
(
    data=,
    id=,
    type=1,
    var=,
    weight=,

    adjacency=,
    predefined=,
    map=,

    minclusters=,
    maxclusters=,
    outlib=work,

    plots_only=no,
    rep_root=c:\temp,
    rep_css=../all_reports.css,
    rep_all=all_reports.html,

    animgif=&rep_root\animgif.gif,
    colorscheme=,
    title=Segmentation of &id based on &var,
    title2=,
    footnote=Based on dataset &data,
    nlevels_detail=200,
    nlevels_axis=20,
    log_id=no,
    log_var=no,
    log_weight=no,
    bubble_plot=1,
    bubble_size=1,
    weight_plot=0,
    debug=no
) / des='Categorizes variable (hierarchical, restriction of levels)';


/* primitive - weighted mean */

%macro _wmean(y1,w1,y2,w2);
((&w1*&y1+&w2*&y2)/(&w1+&w2))
%mend _wmean;


/* primitive - distance */

%macro _xdist(res,i1,i2);
if w(&i1)+w(&i2) ne 0 then do;
    &res=(w(&i1)*y(&i1)**2+s(&i1)+w(&i2)*y(&i2)**2+s(&i2))-
         (w(&i1)+w(&i2))*(%_wmean(y(&i1),w(&i1),y(&i2),w(&i2)))**2-s(&i1)-s(&i2);
end;
else do;
    &res=0;
end;
%mend _xdist;


/* primitive - set element in lower triangular matrix */

%macro _dput(i,j,v);
if (&i)>(&j) then d((&i-2)*(&i-1)/2+(&j))=&v;
else _k=.;
%mend _dput;


/* primitive - get element in lower triangular matrix */

%macro _dget(i,j,v);
if (&i)>(&j) then &v=d((&i-2)*(&i-1)/2+(&j));
else if (&i)<(&j) then do;
    &v=d((&j-2)*(&j-1)/2+(&i));
end;
else &v=.;
%mend _dget;


/* draws one map */

%macro _plot_one_map(data,var,varfmt,id,idfmt,map,step,colorscheme);

%local maxcolor ng hlegend alegend;

%let maxcolor=90;

proc sql;
create table _dinp as
select distinct &var from &data;
reset noprint;
select count(*) into :ng from _dinp;
quit;

proc rank data=_dinp out=_ranked group=%sysfunc(min(&maxcolor,&ng));
var &var;
ranks r&var;
run;

proc sql;
create table _scale as
select r&var,min(&var) as min,max(&var) as max from _ranked group by r&var;

create table _res as
select &id,r&var,&var,'title="'||trim(left(put(&id,&idfmt)))||"&#10; &var= "||put(&var,&varfmt)||' "' as html
from &data,_scale where &var between min and max;

create table _legend as
select r&var,max(&var) as &var from _res group by r&var;
quit;

data _null_;
set _legend ;

/* this sets pattern to color on the blue-to-red scale         */
/* BLUE correspond to VMIN level, RED correspond to VMAX level */

ip=r&var+1;
%if %length(&colorscheme)=0 %then %do;
    ic=360-int((r&var)*250/%sysfunc(min(&maxcolor,&ng)));
%end;
%else %do;
    ic=360-int(%sysfunc(min(&maxcolor,&ng))*(&var-&var_min)/(&var_max-&var_min)*250/%sysfunc(min(&maxcolor,&ng)));
%end;
call execute("pattern" || trim(left(put((ip),4.))) ||"  v=s  color=H" || put(ic,HEX3.) || "AAFF ;");
run;

%if &step>40 or &step=0 %then %do;
   %let hlegend=1;
   %let alegend=2;
%end;
%else %do;
   %let hlegend=2;
   %let alegend=1;
%end;

data _null_;
set _legend end=__finish;
if _n_=1 then do;
    call execute("LEGEND1 label=(h=2 pct ""Sazba"" justify=center position=(top center)) POSITION=(MIDDLE RIGHT OUTSIDE)");
    call execute("across=&alegend shape=bar(&hlegend pct,&hlegend pct) FRAME CFRAME=CXE8E8E8 CBORDER=GRAY value=(h=&hlegend pct ");
end;
call execute('"'||put(round(&var,10),5.0)||'"');
if __finish then call execute(");run;");
run;

proc gmap map=&map data=_res all;
id &id;
choro r&var / discrete coutline=black woutline=1 legend=legend1 html=html description="&var._&step";
run;
quit;

%mend _plot_one_map;


%put;
%put NOTE: Macro MODEL_HIER1 2.3;
%put;

%local i n n1 nn2 colorlist step hsym xvar out tree detail
       id_length id_label id_format
       var_min var_max var_label var_format
       weight_min weight_max weight_label weight_format weight_average;

%let out=&outlib..cluster_&id;
%let tree=&outlib..tree_&id;
%let detail=&outlib..detail_&id;


/* Validating parameters */

%if &data= %then %do;
    %put ERROR: Input dataset is required (DATA=).;
    %goto mexit;
%end;

%if &type ne 0 and &type ne 1 and &type ne 2 and &type ne 3 and &type ne 4 and &type ne 5 %then %do;
    %put ERROR: Valid values for TYPE= option are 0 to 5.;
    %goto mexit;
%end;

%if &id= %then %do;
    %put ERROR: ID variable is required (ID=).;
    %goto mexit;
%end;
%else %if %scan(&id,2) ne %then %do;
    %put ERROR: Only one ID variable is required.;
    %goto mexit;
%end;

%if &var= %then %do;
    %put ERROR: Target variable is required (VAR=).;
    %goto mexit;
%end;
%else %if %scan(&var,2) ne %then %do;
    %put ERROR: Only one target variable is required.;
    %goto mexit;
%end;

%if &weight= %then %do;
    %put ERROR: Weight variable is required (WEIGHT=).;
    %goto mexit;
%end;
%else %if %scan(&weight,2) ne %then %do;
    %put ERROR: Only one weight variable is required.;
    %goto mexit;
%end;

%let plots_only=%qupcase(&plots_only);

%let log_id=%qupcase(&log_id);
%if &log_id ne YES and &log_id ne NO %then %do;
    %put ERROR: Valid values for LOG_ID= option are YES or NO.;
    %goto mexit;
%end;
%if &log_id=YES and &type>=3 %then %do;
    %put WARNING: Scale of X cannot be logaritmic due to TYPE>2.;
    %let log_id=NO;
%end;

%let log_var=%qupcase(&log_var);
%if &log_var ne YES and &log_var ne NO %then %do;
    %put ERROR: Valid values for LOG_VAR= option are YES or NO.;
    %goto mexit;
%end;

%let log_weight=%qupcase(&log_weight);
%if &log_weight ne YES and &log_weight ne NO %then %do;
    %put ERROR: Valid values for LOG_WEIGHT= option are YES or NO.;
    %goto mexit;
%end;

%let debug=%qupcase(&debug);


/* Plots only option */

%if &plots_only=YES %then %do;

    /* Test availability of expected datasets */

    %if %sysfunc(exist(&out))=0 %then %do;
        %put ERROR: Dataset %qupcase(&out) not found while PLOT_ONLY=YES.;
        %goto mexit;
    %end;
    %if %sysfunc(exist(&detail))=0 %then %do;
        %put ERROR: Dataset %qupcase(&detail) not found while PLOT_ONLY=YES.;
        %goto mexit;
    %end;
    %if %sysfunc(exist(&tree))=0 %then %do;
        %put ERROR: Dataset %qupcase(&tree) not found while PLOT_ONLY=YES.;
        %goto mexit;
    %end;

    /* Define range of number of cluster to be plotted */

    %if %length(&minclusters)=0 %then %do;
        proc sql;
        select min(ncluster) into :minclusters from &detail;
        quit;
    %end;
    %if %length(&maxclusters)=0 %then %do;
        proc sql;
        select max(ncluster) into :maxclusters from &detail;
        quit;
    %end;

    proc sql noprint;
    select count(*) into :n from &outlib..&id._sorted;
    quit;
    %if &maxclusters>=&n %then %let maxclusters=%eval(&n-1);
    %goto plot_only;
%end;


/* Test of input datasets */

%if %ds_vars(&data,&id &var &weight)=0 %then %do;
    %put ERROR: Invalid input dataset %qupcase(&data) specified.;
    %goto mexit;
%end;

%if &type=2 %then %do;
    %if %ds_vars(&adjacency,&id)=0 %then %do;
        %put ERROR: Invalid sort definition dataset %qupcase(&adjacency) for TYPE=&type.;
        %goto mexit;
    %end;
%end;

%if &type=4 %then %do;
    %if %ds_vars(&adjacency,&id._1 &id._2)=0 %then %do;
        %put ERROR: Invalid adjacency dataset %qupcase(&adjacency) for TYPE=&type.;
        %goto mexit;
    %end;
    %if %length(&map) ne 0 %then %do;
        %if %ds_vars(&map,&id x y)=0 %then %do;
            %put ERROR: Map definition %qupcase(&adjacency) is invalid for TYPE=&type.;
            %goto mexit;
        %end;
    %end;
%end;

%if &type=5 %then %do;
    %if %length(&adjacency)=0 %then %do;

        /* Derive definition of region joins from the map */

        proc sql;
        create table &id._join as
        select distinct
            case when a.&id>b.&id then b.&id else a.&id end as &id._1,
            case when a.&id>b.&id then a.&id else b.&id end as &id._2
            from &map as a,&map as b where a.x=b.x and a.y=b.y and a.&id ne b.&id;
        quit;
        %let adjacency=&id._join;
    %end;
%end;

%if %length(&predefined) ne 0 and (&type=4 or &type=5) %then %do;

    /* Append predefined joins dataset to the adjacency dataset */

    data &id._join;
    set &adjacency &predefined;
    if &id._1>&id._2 then do;
        __aux=&id._1;
        &id._1=&id._2;
        &id._2=__aux;
    end;
    drop __:;
    run;
    proc sql;
    create table &id._join as
    select distinct a.* from &id._join as a where a.&id._1 ne a.&id._2;
    quit;
    %let adjacency=&id._join;
%end;

/* Save ID format and length to macrovars */

%let id_format=%var_format(&data,&id);
%let id_length=%var_length(&data,&id);
%let id_label=%var_label(&data,&id);

/* Save labels of the Y and W vars */

%let var_format=%var_format(&data,&var);
%let var_label=%var_label(&data,&var);
%let weight_format=%var_format(&data,&weight);
%let weight_label=%var_label(&data,&weight);


/* Create dataset with sorted lists of all ID levels */

proc sql; 
%if %sysfunc(exist(&outlib..&id._sorted)) %then %do;
    drop table &outlib..&id._sorted;
%end;
%if %sysfunc(exist(_sort)) %then %do;
    drop table _sort;
%end;
quit;

%if &type<=3 and &type ne 2 %then %do;
        /*ordinal/nominal category*/
        /* use DS for category class level list */
        proc sql;
            create table _sort (keep=&id &var &weight) as
                select distinct(&id), 
                        put(&id,&id_format) as form_&id,
                        &var,
                        &weight
                from &data
                order by
                %if &type=0 %then %do;
                        &id
                    %end;
                %else %if &type=1 %then %do;
                        form_&id
                    %end;
                %else %if &type=3 %then %do;
                        &var
                    %end;
            ;
        quit;
    %end;
%else %if &type=2 %then %do;
        /*ordinal - specific sort order - just copy provided list */
       data _sort (keep=&id);
            set &adjacency (keep=&id);
       run;

       proc sql;
            create table _sort as
                select a.&id ,
                       b.&var,b.&weight
                from _sort as a, &data as b
                where a.&id=b.&id
        ;
    %end;
%else %if &type=4 or &type=5 %then %do;
        /*restricted nominal or MAP => derive all possible levels of ID from adjacency dataset */
       proc sql;
            create table  _sort (keep=&id) as
                select distinct &id
                from (
                        select &id._1 as &id from &adjacency
                    union 
                        select &id._2 as &id from &adjacency
                    );
            create table _sort as
                select a.&id,b.&var,b.&weight from _sort as a, &data as b where a.&id=b.&id;
    %end;

/* Anyway add index of category level and formatted values to sort list so we get standart dataset for all ID types */

data &outlib..&id._sorted (keep=ncat &id form_&id &var &weight);
    attrib &id label="&id_label " format=&id_format;
    format &var &var_format;
    format &weight &weight_format;
    set _sort;
    form_&id=put(&id,&id_format);
    ncat=_n_;
run;

/* Find out number of class levels for ID */

proc sql noprint;
    select count(*) into :n
    from &outlib..&id._sorted;
quit;
%let n=&n; 
%let n1=%eval(&n+1);

/*==============================================*/
/*setup defaul values of minclusters/maxclusters*/
/*==============================================*/
%if %length(&minclusters)=0  %then %do;
    %let minclusters=1;
%end;
%if %length(&maxclusters)=0  %then %do;
    %let maxclusters=%eval(&n-1);
    /*    limit number of levels to maximum of 300 -if maxclusters was not specified */
    %if &maxclusters>300 %then %do;
        %let maxclusters=300;
    %end;
%end;

/*==============================================*/
/*setup weight_plot by bubble_plot */
%if %length(&weight_plot)=0  %then %do;
   %if &bubble_plot=1 %then %do;
        %let  weight_plot =0;
    %end;
    %else %do;
        %let  weight_plot =1;
    %end;

%end;

/***************************************************************************************/
/* Run clustering algorithm with increase of Sum of Squares (SS) as distance criterium */
/***************************************************************************************/

data 
    &out   (label="Clustering procedure MODEL_HIER1 results " keep= nc--sspct)
    &tree  (label="Dendogram output for &data/&var/&id" keep=_name_ _parent_ _height_)
    &detail(label="Cluster definintion for all steps of MODEL_HIER1"  
            keep=ncluster clusterid &id &var ncat &weight &var._cluster &weight._cluster &weight._cluster_pct sscluster sspctcluster)
;
attrib
    /*out dataset definintion*/
    nc      label="Number of clusters"
    c1      label="Identification of first cluster collapsed in this step"
    c2      label="Identification of second cluster collapsed in this step"
    cid     label="Identification of the created cluster"
    mincrit label="Distance c1 c2 found as minimal in this step"
    yc      label="Mean value of Y in created cluster"
    sc      label="Increment of SS by the new cluster"
    scpct   label="Increment of SS by the new cluster as [%] of total SS"       format=percent8.4
    wc      label="Weight of created cluster"
    wcpct   label="Relative weight of created cluster in [%] of total weight"   format=percent8.4
    ss      label="(Weighted) sum of squares"
    sspct   label="Ratio of SS/total SS in [%]"                                 format=percent8.4
;
attrib
    /*detail dataset definintion*/
    ncluster             label="Number of clusters"
    clusterid            label="Cluster identification"
    &var._cluster        label="&var_label in cluster"
    &weight._cluster     label="&weight_label in cluster"
    &weight._cluster_pct label="% of total &weight_label in cluster"     format=percent8.4
    sscluster            label="Sum of squared in cluster"
    sspctcluster         label="% of total sum of squared in cluster"    format=percent8.4
    ncat                 label="Level # of &id_label"
;

/*-----------------------*/
/*array definintions*/
/*-----------------------*/

array x(&n) &id_length   
                _temporary_;   /* the original set ID levels */ 
array y0(&n)    _temporary_;   /* the original Y values (for detail output) */ 
array w0(&n)    _temporary_;   /* the original w values (for detail output) */ 
array c0(&n)    _temporary_;   /* cluster identification for original set of levels 
                                    c0(i) >0 => then it represent number of original level in category ID
                                    c0(i) <0 => cluster created in previous steps */

/* All arrays contains meanfull values only for elements 1 to NC (number of clusters in actual step) */
array c(&n)     _temporary_;   /* cluster identification 
                                    c(i) >0 => then it represent number of original level in category ID
                                    c(i) <0 => cluster created in previous steps */
array y(&n)     _temporary_;   /* interval variable by which segmentation is performed */
array s(&n)     _temporary_;   /* sum of squares within cluster */
array w(&n)     _temporary_;   /* weight of bins/clusters */ 

/*distance matrix definition*/
%if &type<=2 and
    /*if there is dataset with predefined joins it is necessary to run nominal */
    /*with distance matrix*/
    (&type=3 and %length(&predefined)=0) %then %do;
    /*distance array for ordinal and nominal variables*/
    array d(&n)  _temporary_;
%end;
%else %do;
    /* definintion of triagunal matrix    */
    /* distance matrix for nominal variables*/
    %let nn2=%eval(&n*(&n-1)/2);
    array d(&nn2)  _temporary_;
%end;

attrib
    ytot label="Average of Y over all observations"
    wtot label="Weight of Y over all observations"
    stot label="Std. deviation of Y over all observations"
;

/*================================*/
/*read input data set and calculate sums aver ID levels*/
/*================================*/
if _n_=1 then do;
    /*-------------------------------*/
    /* read ordered lists of bins for all categories*/
    /*-------------------------------*/
    do i= 1 to &n;
        set &outlib..&id._sorted;
        x(i)=&id;
    end;
    /*-------------------------------*/
    /*   reset Y,S,W,C  arrays*/
    /*-------------------------------*/
    do i= 1 to &n;
        y(i)=0;
        s(i)=0;
        w(i)=0;
        c(i)=i;
    end;

    /* prepare distance matrixes for nominal variables - set d(i,j) to missing if i,j join is not allowed */
    %if &type=3 and %length(&predefined) ne 0 %then %do;
    /* reset distance matrix for nominal category with predefined joins = all joins allowed */
        do i= 1 to &n;
            do j= 1 to i-1;
                %_dput(i,j,0);
            end;
        end;
      %end;
    %else %if &type>3 %then %do;
    /* reset distance matrix for resticted nominal category missing= join not allowedc */
        do i= 1 to &n;
            do j= 1 to i-1;
                %_dput(i,j,.);
            end;
        end;
        /* read adjency matrix*/
        do until (fin2);
            set &adjacency (keep=&id._1 &id._2 ) end=fin2;
            do i= 1 to &n;
                if &id._1=x(i) then leave;
            end;
            do j= 1 to &n;
                if &id._2=x(j) then leave;
            end;
            if i>j then do;
                %_dput(i,j,0);
            end;
            else do;
                %_dput(j,i,0);
            end;
        end;
    %end;
end;

/*================================*/
/*main cycle of the datastep -reading of data*/
/*================================*/
/* read input data set */
set &data (obs=&n) end=__finish;

do i= 1 to &n;
    if &id=x(i) then leave;
end;

/*  add observation into proper place in Y,S,W arrays */
y(i)+&var*&weight;
w(i)+&weight;
/*================================*/
if __finish;
/* end of the datatep cycle - continue here after whole data was read*/
/*================================*/
/*-------------------------------*/
/* calculate weighted average of Y over all bins */
/*-------------------------------*/
ytot=0;
stot=0;
wtot=0;
do i= 1 to &n;
   /* normalize recalculate Y S to sum of weight in bin*/
/*    put y(i)= w(i)= ;*/
    ytot+y(i);
    wtot+w(i);
    if w(i) ne 0 then do; 
        y(i)=y(i)/w(i);
    end;
    stot+y(i)**2*w(i);
    /*    save original values to array used for detail output*/
    y0(i)=y(i);
    w0(i)=w(i);
    c0(i)=i;
end;
ytot=ytot/wtot;
stot=stot-wtot*ytot**2;
sqrtstot=sqrt(stot);

/*------------------------------*/
/*calculate initial distances of adjacent levels */
/*------------------------------*/
%if &type<=2 or (&type=3 and %length(&predefined)=0) %then %do;
    /*ordinal/nominal  category*/
    do i=1 to &n-1;
        %_xdist(d(i),i,i+1); 
    end;
%end;
%else %do;
    /*map categories + nominal with predefined dataset*/
    do i=1 to &n;
        do j=1 to i-1;
            %_dget(i,j,dij)
            if dij ne . then do;
                %_xdist(dij,i,j);
                %_dput(i,j,dij);
                %_dget(i,j,dij)
            end;
        end;
    end;

    /*------------------------*/
    %if &debug=YES %then %do;
        nc=&n;
        array q(&n);
        do i=1 to nc;
          do j=1 to i;
            %_dget(i,j,dij);
            q(j)=dij;
          end;
          put i= 4.0 @;
          do j=1 to i;
            put q(j) e12.4 @;
          end;
          put ;
        end;
    %end;

%end;

/*=================================*/
/*run hierachical clustering */
/*nc = number of clusters */
ss=0;
do nc=&n-1 to 1 by -1;

    /*------------------------*/
    %if &debug=YES %then %do;
        put / 'nc=' nc 3.0 +1 @;
        do j=1 to nc+1;
            put j 5.0 +8 @;
        end;
        put;

        /*------------------------*/

        put / 'nc=' nc 3.0 +1 @;
        do j=1 to nc+1;
            put j 5.0 +8 @;
        end;
        put;

        put 'C' +3 @;
        do i=1 to nc+1;
        put c(i) 12.3 +1 @;
        end;
        put;

        put 'Y' +3 @;
        do i=1 to nc+1;
        put y(i) 12.3 +1 @;
        end;
        put;

        put 'W' +3 @;
        do i=1 to nc+1;
            put w(i)  12.3 +1 @;
        end;
        put;

        put 'S' +3 @;
        do i=1 to nc+1;
        put s(i) 12.3 +1 @;
        end;
        put;

        put 'V' +3 @;
        do i=1 to nc+1;
            put Y(i) 12.3 +1 @;
        end;
        put;

        put 'C0' +3 @;
        do i=1 to &n;
            put c0(i) 12.3 +1 @;
        end;
        put;
        put 'X' +3 @;
        do i=1 to &n;
            put x(i) +1 @;
        end;
        put;


    %end;
    /*--------------------*/
    if not finpredef and not missing(trim("&predefined")) then do;
    /*=================================*/
    /*  process predefined clusters defined in */
    /*=================================*/
        set &predefined end=finpredef;

        /* find position of the level2 in the array of original names */
        do ii1= 1 to &n;
            if &id._1=x(ii1) then leave;
        end;
        do i1=1 to nc;
            if c(i1)=c0(ii1) then leave;
        end;

        do ii2= 1 to &n;
            if &id._2=x(ii2) then leave;
        end;
        do i2=1 to nc;
            if c(i2)=c0(ii2) then leave;
        end;


        /* define distance*/
        %if &type<=2 or (&type=3 and %length(&predefined)=0) %then %do;
            /*ordinal/nominal categories*/
            mincrit=d(i1);
        %end;
        %else %do;
            /*map categories*/
            %_dget(i2,i1,dij);
            mincrit=dij;
        %end;

    end;

/* find closest clusters */

    else do;
        mincrit=.;
        %if &type<=3 %then %do;
            /*ordinal/nominal categories*/
            do i=1 to nc;
                if d(i)<mincrit or mincrit=. then do;
                    i1=i;
                    i2=i+1;
                    mincrit=d(i);
                end;
            end;
        %end;
        %else %do;
            /*map categories*/
            do i=1 to nc+1;
                do j=1 to i-1;
                    %_dget(i,j,dij);
                    if dij ne . then do;
                        if dij<mincrit or mincrit=. then do;
                            i1=j;
                            i2=i;
                            mincrit=dij;
                        end;
                    end;
                end;
            end;
        %end;
    end;

    /* set identification of the newly created cluster */
    %if &debug=YES %then %do;
        put / nc= "collapsing " i1= i2= mincrit=;
    %end;
    cid=-nc;
    c1=c(i1);
    c2=c(i2);
    /* set new cluster identification in "long" array (replace identificator c(I1) or c(I2) by -NC)*/
    do i=1 to &n;
        if c0(i)=c1 or c0(i)=c2 then do;
            c0(i)=cid;
        end;
    end;
    /*=================================*/
    sc=mincrit; 
    scpct=sc/stot;
    /* recalculate Y S W for new cluster created when we join c1 and c2 */
    wc=(w(i1)+w(i2));
    wcpct=wc/wtot;
    if w(i1)+w(i2) ne 0 then do;
        yc=%_wmean (y(i1),w(i1),y(i2),w(i2));
       end;
    else do;
        yc=0;
       end;
    %if &debug=YES %then %do;
        put  'joined ' c1= c2= y(i1)= y(i2)= yc=;
    %end;
    /*----------------*/
    /* output dendoram control file */
    _parent_=cid;
    _name_=c1;
    _height_=sspct;
    if c1>0 then _height_=0;
    output &tree;
    _name_=c2;
    _height_=sspct;
    if c2>0 then _height_=0;
    output &tree;
    /*----------------*/
    /* re-calculate sum of squares corresponding to nc clusters */
    ss+sc;  
    sspct=ss/stot;
    %if &debug=YES %then %do;
        put nc= 'new cluster ' yc= wc=  sc= scpct= ss= sspct= /;
    %end;
    /* output newly formed cluster configuration */
    output &out;
    /*----------------*/
    /*  update clusters/bin identification */
    c(i1)=cid;
    /*  update I1-th array element of Y,S,W arrays and shifting all array elements above I2*/
    w(i1)=wc;
    y(i1)=yc;
    s(i1)=sc+s(i1)+s(i2);
    do i=i2 to nc;
        y(i)=y(i+1);
        s(i)=s(i+1);
        w(i)=w(i+1);
        c(i)=c(i+1);
    end;

    %if &debug=YES %then %do;
        put 'S' +3 @;
        do i=1 to nc+1;
        put s(i) 12.3 +1 @;
        end;
        put;
    %end;

/************************************************/
/* generate detail cluster configuration output */
/************************************************/

    if nc>=&minclusters and nc<=&maxclusters then do;
        ncluster=nc;
        do ic=1 to nc;
            clusterid=c(ic);
            &var._cluster =y(ic);
            &weight._cluster =w(ic);
            &weight._cluster_pct =w(ic)/wtot;

            sscluster=s(ic);
            sspctcluster=s(ic)/stot;
            do i=1 to &n;
                if c0(i)=clusterid  then do;
                    ncat=i;
                    &id=x(i);
                    &var=y0(i);
                    &weight=w0(i);
                    output &detail;
                end;
            end;
        end;
    end;

/********************/
/* update distances */
/********************/

    %if &type<=3 %then %do;
    /* ordinal/nominal category*/
    /*----------------*/
        do i=i2 to nc;
            d(i)=d(i+1);
        end;
        if i1>1 then do;
            %_xdist(d(i1-1),i1-1,i1); 
        end;
        %_xdist(d(i1),i1,i1+1);  
    %end;
    %else %do;

/******************/
/* map categories */
/******************/

/* mark elements adjoin to I1,I2 cluster */

       do i=1 to nc+1;
            %_dget(i,i1,dii1);
            %_dget(i,i2,dii2);
            if dii1 ne . or dii2 ne .
            then do;
                if i>i1 then do;
                   %_dput(i,i1,0);
                   end;
                else do;
                   %_dput(i1,i,0);
                end;
            end;
        end;

/* shift distance array elements*/

        do i=i2 to nc;
            do j= 1 to i2-1;
                %_dget(i+1,j,dij);
                %_dput(i,j,dij);
            end;
        end;
        do j=i2 to nc;
            do i=j+1 to nc;
                %_dget(i+1,j+1,dij);
                %_dput(i,j,dij);
            end;
        end;

    do j=1 to nc;
        %_dget(i1,j,di1j);
        if di1j ne . then do;
            %_xdist(di1j,i1,j); 
            if i1>j then do;
                    %_dput(i1,j,di1j); 
                end;
            else do;
                    %_dput(j,i1,di1j); 
                end;
        end;
    end;

        %if &debug=YES %then %do;
            do i=1 to nc;
              do j=1 to i;
                %_dget(i,j,q(j));
              end;
              put i= 4.0 @;
              do j=1 to i;
                put q(j) e12.4 @;
              end;
              put ;
            end;
        %end;


    %end;
/*=================================*/
end; *over nc = cluster identification;
/*=================================*/

/* output last node to dendoram control file*/
_parent_=.;
_height_=sspct;
_name_=-1;
output &tree;
stop;
run;

%plot_only:

/******************************************************************/
/*  start outputs = delete graphical catalog & setup ODS/GOPTIONS */
/******************************************************************/

proc datasets nolist;
    delete gseg / memtype=catalog;
run;

%let colorlist=BLACK RED GREEN BLUE CYAN MAGENTA GRAY PINK ORANGE BROWN;

goptions reset=all;

goptions 
    device=gif
    ftext="Arial" htext=3 pct htitle=5 pct
    colors=(&colorlist)
;
/*set size of the points for different number of levels*/
%if &n<&nlevels_detail   %then %do;
    %let hsym=2;
%end;
%else %do;
    %let hsym=0.1;
%end;

%directory_create(&rep_root);

ods listing close;
ods html path="&rep_root" (url=none)
         frame="%sysfunc(lowcase(&id)).html"
         body="%sysfunc(lowcase(&id))_b.html" (title="&title")
         contents="%sysfunc(lowcase(&id))_c.html"
         stylesheet=(url="&rep_css");
title "Sum of Squares";
%if "&title2" ne "" %then %do;
    title2 "&title2";
%end;
symbol1 i=j c=blue v=dot h=&hsym pct;
axis4
    label=(angle=90 r=0 )
    order =(0 to 1 by 0.05)
;
proc gplot data=&out;
    format sspct percent6.0;
    plot sspct*nc /vaxis=axis4 ;
run;

/****************************************/
/*prepare format category level numbers */
/****************************************/

data _cntl;
    set &outlib..&id._sorted end=__finish;
    label=form_&id;
    start=ncat;
    fmtname='ncat';
    type='n';
    output;
run;
proc format cntlin=_cntl;
run;

/*************************/
/*print clustering steps */
/*************************/

%if &n<&nlevels_detail  %then %do;
    title h=2 f='arial' 'Clusters';
%if "&title2" ne "" %then %do;
    title2 "&title2";
%end;
    proc print label data=&out;
        id nc;
        format c1 ncat. c2 ncat.;
        var c1 c2 cid sspct wcpct;
    run;
%end;

%if &n<50 %then %do;

/*******************/
/* plot dendrogram */
/*******************/

    proc sql;
        create table _tree as
            select *
            from &tree as a left join &outlib..&id._sorted as b on _name_=ncat;

    data _tree;
        set _tree;
        if missing(ncat) then do;
            form_&id=put(_name_,8.0);
        end;
    run;

    title "Dendrogram";
%if "&title2" ne "" %then %do;
    title2 "&title2";
%end;
    proc tree data=_tree horizontal;
        id form_&id;
    run;
%end;


%if &type>=4 or (&type=3 and %length(&predefined) ne 0)%then %do;
/*derive correct order for restricted and map dataset */
/*order is similar to the orde of category level in dendrogram */
    proc sort data=&out out=_tree;
        by nc;
    run;

    data _order (keep =ncat ncat_sort );
    array _ncat_sort(&n) (&n*0);
    retain kmax;
    set _tree end=__finish;
    if _n_=1 then do;
        _ncat_sort(1)=c1;
        _ncat_sort(2)=c2;
        kmax=2;
    end;
    else do;
        /*search for parent nore*/
        do k1=1 to kmax;
            if _ncat_sort(k1)=cid then leave;
        end;
        /*insert this left/right nodes to ordered list (array)*/
        _ncat_sort(k1)=c1;
        ck1=c2;
        kmax+1;
        do k=k1+1 to kmax;
            ck2=_ncat_sort(k);
            _ncat_sort(k)=ck1;
            ck1=ck2;
        end;
    end;
    if __finish;
    do ncat_sort=1 to &n;
        ncat=_ncat_sort(ncat_sort);
        output;
    end;
    run;

    /*merge proper order NCAT_Order with sorted copy od data*/
    proc sort ;
        by ncat;
    run;
    data &outlib..&id._sorted;
      label ncat_sort ="&id_label ";
      merge &outlib..&id._sorted _order;
      by ncat;
    run;
    proc sort ;
        by ncat_sort;
    run;
%end;
%else %do;
    /*just add ncat_order equal to ncat*/
    data &outlib..&id._sorted;
      label ncat_sort ="&id_label ";
      set &outlib..&id._sorted;
      ncat_sort=ncat;
    run;
%end;

/************************************************/
/* prepare format based on sorted category list */
/************************************************/

data _cntl (keep= start label type fmtname);
    set &outlib..&id._sorted end=__finish;
    label=form_&id;
    start=ncat_sort;
    fmtname='sncat';
    type='n';
    output;
    if __finish;
        label='';
        start=0;
        output;
        start=&n+1;
        output;
run;
proc format cntlin=_cntl;
run;

/******************************************************/
/* print original data categories sorted as on graphs */
/******************************************************/

%if &n<&nlevels_detail  %then %do;
    title h=2 f='arial' "Input data sorted as on graphs";
%if "&title2" ne "" %then %do;
    title2 "&title2";
%end;
    proc print label data=&outlib..&id._sorted;
        id ncat_sort;
        var &id &var &weight;
    run;
%end;

/**************************************************************/
/* find out stepmax based on minprintcatlist or print_flag at */
/**************************************************************/

ods html newfile=proc;

/*define set of symbols*/
symbol1;symbol2;symbol3;
axis1   label=( angle=90 r=0 )
    %if &log_var=YES %then %do;
        logbase=10
        logstyle=expand 
    %end;
;
axis2   label=( angle=90 r=0 )
    %if &log_weight=YES %then %do;
        logbase=10
        logstyle=expand 
    %end;
;
/*define axis to be used with ncat_sort format*/
axis3
    label=( angle=0 r=0 )
    %if &n<&nlevels_axis %then %do;
        value=(angle=45 r=0 )
        order =(0 to &n1 by 1 )
    %end; 
    %else %do;
        %if &log_id=NO %then %do;
           order =(0 to %sysevalf(5*(%sysfunc(int(&n/5))+1)) by 5 )
/*order =(0 1 to &n1 by %eval(&n1/10) &n )*/
            value =(angle=90 r=0 )
        %end;
        %else %do;
            logbase=10
            logstyle=expand 
        %end;
    %end; 
;

proc sql noprint; 
select avg(&weight),min(&weight),max(&weight) into :weight_average,:weight_min,:weight_max from &data;

/*find max and min values of the &var*/
proc sql noprint;
select min(&var),max(&var) into :var_min,:var_max from &data;

/*============================*/
/*main cycle over minclusters maxclusters*/
%do step=&minclusters  %to &maxclusters; 
/*============================*/
proc sql;
        create table _step as
        select a.*,
               b.ncat_sort
        from &detail (where=(ncluster=&step)) as a , 
             &outlib..&id._sorted as b
        where a.ncat=b.ncat              
        order by clusterid, ncat_sort 
    ;
     run;

    /*    generate clicable maps that correspond to _STEP */
    %if &type=5 or (&type=4 and %length(&map) ne 0 ) %then %do;
        title "Clusters averages in category &id /&step clusters";
%if "&title2" ne "" %then %do;
    title2 "&title2";
%end;
        %_plot_one_map(_step,&var._cluster,&var_format,&id,&id_format,&map,&step,&colorscheme);
    %end;
     
    /*generate annotate datasets for category clusters*/
    data _step (keep= clusterid &var._cluster &weight._cluster &weight._cluster_pct &id ncat ncat_sort &var &weight html)
         _yano (keep=x y xsys ysys color function size style line html  angle rotate hsys)
         _wano (keep=x y xsys ysys color function size style line html)
    ;
           set _step end=__finish;
           by clusterid notsorted ;

           retain x1 .; 
           length color $8 ; retain color ;
           length icolor 8 ; retain icolor 0;
           length html $200;

            /* html flyovers*/
           %if &n<&nlevels_detail   %then %do;
               html='title="' || 
                    "&id=" || compress(put(&id,&id_format)) || ' &#10; ' || 
                    "&var=" || compress(put(&var,&var_format)) || ' &#10; ' || 
                    "&weight=" || compress(put(&weight,&weight_format)) ||  '"';
            %end;
           output _step;
 
          if first.clusterid  and not last.clusterid  then do;
                x1=%if &log_id=NO %then ncat_sort;%else &id;;
          end;

           /*  annotate defaults*/
           xsys='2';ysys='2'; function="move"; size=1; line=1;style='solid';   
           /*renormalize cluster weights to its width */
           if first.clusterid then do;
                icolor+1;
                color=scan("&colorlist",icolor); 
                if color='' then do; icolor=1; color=scan("&colorlist",icolor); end;
           end;

            /*  generate bubble plots*/
           %if &bubble_plot=1 %then %do;
                function="pie"; 
                x=%if &log_id=NO %then ncat_sort;%else &id;; 
                y=&var; 
                angle=0;
                rotate=360;
                hsys='1';
                size=&bubble_size*(sqrt(&weight))/(sqrt(&weight_max))/min(&n/100,5);
                if size<0.1 then size=0.1;
                %if &log_var=YES %then %do; 
                    if &var>0 then 
                %end;
                        output _yano;
                size=1;
           %end;

            /*         cluster made by the single level*/
           if first.clusterid  and last.clusterid  and &n<&nlevels_axis then do;
                 function="move"; 
                x=%if &log_id=NO %then ncat_sort;%else &id;-0.3; 
                y=&var._cluster; 
                output _yano;
                y=&weight._cluster; 
                output _wano;
                function="draw";
                x=%if &log_id=NO %then ncat_sort;%else &id;+0.3; 
                y=&var._cluster; 
                output _yano;
                y=&weight._cluster; 
                output _wano;
          end;

        /* bar generated to annotate dataset when end of segment */
           if last.clusterid  and not first.clusterid  then do;
                x=x1-0.3; 
                html='';
                function="move"; 
                y=&var._cluster  ; 
                output _yano;
                wc=&weight._cluster/(ncat_sort-x1+1);
                y=wc;
                output _wano;

                if -clusterid=&step then do;
                    /*eventually change next line to mark segment that was created in this step*/
                    barsize=1.3;
                end;
                else do;
                    barsize=1.3;
                end;

                ysys=7;
                y=-barsize  ; 
                output _yano;
                output _wano;


                %if &step<&nlevels_detail %then %do;
                    html='title="' || 
                    "Cluster of &id # " || compress(put(clusterid,5.)) || ' &#10; ' || 
                    "&var=" || compress(put(&var._cluster,&var_format)) || ' &#10; ' || 
                    "&weight(pct of total)=" || compress(put(&weight._cluster_pct,percent8.4)) || ' &#10; ' || 
                    "&weight(in cluster)=" || compress(put(&weight._cluster,&weight_format)) || '"';
               %end;
                x=%if &log_id=NO %then ncat_sort;%else &id;+0.3;
                function="bar";
                y=barsize;
                output _yano;
                output _wano;
           end;
        run;


        ods proclabel="&var - Graph for &step clusters";

        %if &log_id=NO %then %do;
            %let xvar=ncat_sort;
        %end;
        %else %do;
            %let xvar=&id;
        %end;

        %if &bubble_plot ne 1 %then %do;
            symbol1 v=dot  i=none repeat=1 h=&hsym pct;
            symbol2 v=plus i=none repeat=1 h=&hsym pct;
            symbol3 v=star i=none repeat=1 h=&hsym pct;
        %end;
        %else %do;
            symbol1 v=dot  i=none repeat=1 h=0.1 pct;
            symbol2 v=plus i=none repeat=1 h=0.1 pct;
            symbol3 v=star i=none repeat=1 h=0.1 pct;
        %end;

        title "Bins/clusters averages in category &id /&step clusters";
%if "&title2" ne "" %then %do;
    title2 "&title2";
%end;
        proc gplot data=_step  annotate=_yano ;
            %if &log_var=YES %then %do;
                format &var e7.0;
            %end;

            format ncat_sort sncat.;   
            plot &var*&xvar=clusterid /haxis=axis3 vaxis=axis1 
                description="Cluster Average" name="ybin&step" 
            /* html flyovers*/
               %if &n<&nlevels_detail and &bubble_plot ne 1 %then %do;
                    html=html
               %end;
               ;
        run;

        %if  &weight_plot=1 %then %do;

                symbol1 v=dot  i=none repeat=1 h=&hsym pct;
                symbol2 v=plus i=none repeat=1 h=&hsym pct;
                symbol3 v=star i=none repeat=1 h=&hsym pct;

                ods proclabel="&weight Graph for &step clusters";
                title "Weights &weight in category &id / &step clusters";
%if "&title2" ne "" %then %do;
    title2 "&title2";
%end;
                proc gplot data=_step annotate=_wano;
                    %if &log_weight=YES %then %do;
                        format &weight e7.0;
                    %end;
                    format ncat_sort sncat.;   
                    plot  &weight*&xvar=clusterid /haxis=axis3 vaxis=axis2 html=html
                        description="Cluster Weight" name="wbin&step";
                run;

        %end;

        %if &type<=2  %then %do;
            proc sort data=_step ;
            by ncat;
            run;
        %end;

title h=2 f='arial' "Table for &step clusters";
%if "&title2" ne "" %then %do;
    title2 "&title2";
%end;

%if &n<&nlevels_detail  %then %do;
    proc report data=_step nowd contents="&var Table for &step clusters" 
    style(summary)=[background=yellow font_weight=bold];
    column clusterid 
           &var._cluster
           &weight._cluster
           &weight._cluster_pct
           ncat 
           &id
           &var 
           &weight 
    ;
    define Clusterid / order order=data;
    define &var._cluster / order format=comma15.2;
    define &weight._cluster / order format=comma15.2;
    define &weight._cluster_pct / order;
    define ncat / order noprint;
    define &var / mean format=comma15.2;
    define &weight / mean format=comma18.2;
    run;
%end;
%else %do;
    proc sql;
    create table _cstep as
    select distinct
        clusterid,
        &var._cluster,
        &weight._cluster,
        &weight._cluster_pct,
        min(ncat_sort) as mincat label="Category level from ",
        max(ncat_sort) as maxcat label="Category level to"
    from _step group by clusterid order by mincat;
    quit;

    title h=2 f='arial' 'Clusters - step &step';
%if "&title2" ne "" %then %do;
    title2 "&title2";
%end;
    proc print data=_cstep label;
    id clusterid;
    format mincat ncat.;
    format maxcat ncat.;
    run;
%end;

%end;  *loop over STEP;

%if &type=5 or (&type=4 and %length(&map) ne 0 ) %then %do;
    title "Map with the original data";
%if "&title2" ne "" %then %do;
    title2 "&title2";
%end;
    %_plot_one_map(&data,&var,&var_format,&id,&id_format,&map,0,&colorscheme);
%end;

/* print original data at the end of report if intput is too long */

%if &n>=&nlevels_detail  %then %do;
    title "Input data sorted as on graphs";
%if "&title2" ne "" %then %do;
    title2 "&title2";
%end;
    proc print label data=&outlib..&id._sorted;
        id ncat_sort;
        var &id &var &weight;
    run;
%end;

ods html close;
ods listing;

/************************/
/*generate animated GIF */
/************************/

%if %length(&animgif) ne 0 and &type>=4 %then %do;

    goptions reset=all;
    filename animout "&animgif";
    goptions dev=gifanim gsfname=animout gsfmode=replace notransparency delay=100;

    goptions hsize=8in vsize=6in hpos=83 vpos=43;

    %do i=&maxclusters %to &minclusters %by -1;
        %_plot_one_map(&detail (where=(ncluster=&i)),&var._cluster,&var_format,&id,&id_format,&map,&i,&colorscheme);
        goptions gsfmode=append;
    %end;
    %do i=&minclusters %to &maxclusters %by 1;
        %_plot_one_map(&detail (where=(ncluster=&i)),&var._cluster,&var_format,&id,&id_format,&map,&i,&colorscheme);
        goptions gsfmode=append;
    %end;
    goptions gepilog='3B'x;

%end;

%mexit:

%mend model_hier1;
