/*

-------------------------------------------------------
Makro: MODEL_RANK
Nzev: Rozdlen pozorovn do zadanho potu interval
Verze: 2.21
Autor: jan.ruzicka@cze.sas.com
-------------------------------------------------------

-----
POPIS
-----

Dl pozorovn z tabulky DATA= podle hodnot spojit promnn (VAR=) do zadanho potu
interval (volba GROUPS=).

Vstup:

data - tabulka uren parametrem DATA=
     - mus obsahovat promnnou definovanou parametrem VAR=

Vstup:

out            - pvodn tabulka obsahujc navc promnnou s slem intervalu 0 a groups-1
ranks          - tabulka popisujc jednotliv intervaly
fmtout.fmtname - formt pevdjc promnnou (VAR=) na slo intervalu (pokud je poadovn)
                 (msto sla intervalu me formtovan hodnota bt popisem intervalu)

------
SYNTAX
------

%model_rank
(
    data=,    *** vstupn dataset
    var=,     *** podle jak promnn dlit do interval
    groups=,  *** poet interval
    out=,     *** vstupn tabulka
    outvar=,  *** nzev vstupn promnn s slem intervalu
    ranks=,   *** vstupn tabulka s popisem interval
    fmtout=,  *** knihovna i katalog, kam uloit vstupn formt
    fmtname=, *** nzev vstupnho formtu (nen-li, pouije se VAR=)
    fmtvalue= *** formtovan hodnota (RANK, INTERVAL)
)

-------------
HISTORIE ZMN
-------------

21.05.2002  Verze 1.0
24.09.2002  Verze 2.0
            - pravy algoritmu
12.11.2007  Verze 2.21

--------
PKLADY
--------

* rozdl pozorovn z tabulky SASHELP.CLASS do 3 interval podle velikosti promnn WEIGHT;
* sla interval 0 a groups-1 budou uloeny v promnn WEIGHT_RANK;
* vstupem bude tabulka WEIGHT a popis interval v tabulce WEIGHT_INT;

%model_rank
(
    data=sashelp.class,
    var=weight,
    groups=3,
    out=weight,
    ranks=weight_int
)

--------
POZNMKY
--------

Pipomnky a npady k vylepen vtny.

*/

%macro model_rank
(
    data=,
    var=,
    groups=,
    out=,
    outvar=,
    ranks=,
    fmtout=,
    fmtname=,
    fmtvalue=
) / des='Categorizes variable and creates format (quantile method)';

%put;
%put NOTE: Macro MODEL_RANK 2.21;
%put;

%local i dsid vartype varfmt;
%let i=0;

%if &data= %then %do;
    %put ERROR: Input dataset is required (DATA=).;
    %goto mexit;
%end;

%if &var= %then %do;
    %put ERROR: Variable is required (VAR=).;
    %goto mexit;
%end;
%else %if %scan(&var,2) ne %then %do;
    %put ERROR: Only one variable is required.;
    %goto mexit;
%end;

%let dsid=%sysfunc(open(&data));
%if &dsid %then %do;
    %if %sysfunc(varnum(&dsid,&var))=0 %then %do;
        %put ERROR: Variable %upcase(&var) is not available.;
        %goto mexit;
    %end;
    %else %do;
        %let vartype=%sysfunc(vartype(&dsid,%sysfunc(varnum(&dsid,&var))));
        %if &vartype=C %then %do;
            %put ERROR: Variable %upcase(&var) is not numeric.;
            %goto mexit;
        %end;
        %else %let varfmt=%sysfunc(varfmt(&dsid,%sysfunc(varnum(&dsid,&var))));
    %end;
    %let dsid=%sysfunc(close(&dsid));
%end;
%else %do;
    %put ERROR: Input dataset %upcase(&data) is not available.;
    %goto mexit;
%end;

%if &groups= %then %do;
    %put ERROR: Number of groups is required (GROUPS=).;
    %goto mexit;
%end;
%if %sysfunc(translate(1&groups,"          ","0123456789")) ne %then %do;
    %put ERROR: GROUPS= option expects number.;
    %goto mexit;
%end;
%if %scan(&groups,2) ne %then %do;
    %put ERROR: GROUPS= option expects only one number.;
    %goto mexit;
%end;

%if &out= %then %do;
    %put ERROR: Output dataset is required (OUT=).;
    %goto mexit;
%end;
%else %if %scan(&out,2) ne %then %do;
    %put ERROR: More than one output table defined.;
    %goto mexit;
%end;

%if &outvar= %then %do;
    %let outvar=&var._rank;
    %put NOTE: Rank variable will be %upcase(&outvar).;
%end;
%else %if %scan(&outvar,2) ne %then %do;
    %put ERROR: More than one output variable defined.;
    %goto mexit;
%end;

%if %scan(&ranks,2) ne %then %do;
    %put ERROR: More than one ranks table defined.;
    %goto mexit;
%end;

%if &fmtout= %then %if &fmtname ne %then %do;
    %put ERROR: FMTNAME= option is valid only with FMTOUT= defined.;
    %goto mexit;
%end;
%if &fmtname= %then %do;
    %let fmtname=&var;
    %let i=1;
%end;
%if %scan(&fmtname,2) ne %then %do;
    %put ERROR: More than one output format defined.;
    %goto mexit;
%end;
%else %do;
    %let fmtname=%upcase(&fmtname);
    %if %length(&fmtname)>8 %then %let i=1;
    %let fmtname=%substr(&fmtname,1,%sysfunc(min(8,%length(&fmtname))));
    %if &i %then %put NOTE: Output format name will be &fmtname..;
%end;

%let fmtvalue=%upcase(&fmtvalue);
%if &fmtvalue ne %then %if &fmtout= %then %do;
    %put ERROR: FMTVALUE= option is valid only with FMTOUT= defined.;
    %goto mexit;
%end;
%if &fmtvalue= %then %let fmtvalue=RANK;
%if &fmtvalue ne RANK %then %if &fmtvalue ne INTERVAL %then %do;
    %put ERROR: Valid values for FMTVALUE= option are RANK or INTERVAL.;
    %goto mexit;
%end;

proc rank data=&data out=&out groups=&groups;
var &var;
ranks &outvar;
run;

%if &ranks= %then %if &fmtout= %then %goto mexit;

data %if &ranks ne %then &ranks (keep=_rank_ _interval_ &outvar._start &outvar._end
                                 rename=(_rank_=&outvar));
     %if &fmtout ne %then
     _&fmtname (keep=_&fmtvalue._ &outvar._start &outvar._end fmtname type
                rename=(_&fmtvalue._=label &outvar._start=start &outvar._end=end));;
attrib _rank_ label='';
array _r_(0:&groups,2) _temporary_;
do until(_finish_);
    set &out (keep=&var &outvar rename=(&outvar=_rank_)) end=_finish_;
    if _rank_=. then _rank_=&groups;
    if _r_(_rank_,1)=. then _r_(_rank_,1)=&var;
    else _r_(_rank_,1)=min(_r_(_rank_,1),&var);
    _r_(_rank_,2)=max(_r_(_rank_,2),&var);
end;
type='N';
fmtname="&fmtname";
do _i_=&groups,0 to %eval(&groups-1);
    &outvar._start=_r_(_i_,1);
    &outvar._end=_r_(_i_,2);
    if &outvar._start ne . then do;
        if &outvar._start ne &outvar._end then
            %if &varfmt ne %then %do;
            _interval_=compress(put(&outvar._start,&varfmt))||' - '||
                       compress(put(&outvar._end,&varfmt));
        else _interval_=compress(put(&outvar._start,&varfmt));
            %end;
            %else %do;
            _interval_=compress(put(&outvar._start,best.))||' - '||
                       compress(put(&outvar._end,best.));
        else _interval_=compress(put(&outvar._start,best.));
            %end;
        if _i_ ne &groups then _rank_=_i_;
        else _rank_=.;
        output;
    end;
end;
run;

%if &fmtout ne %then %do;
    proc format lib=&fmtout cntlin=_&fmtname;
    run;
    proc datasets lib=work mt=(data view) nolist;
    delete _&fmtname;
    quit;
%end;

%mexit:

%mend model_rank;
